import numpy as np
from PIL import Image
from Crypto.Cipher import ChaCha20

def split_quadrants(matrix):
    res = np.split(matrix, 2, axis=1)
    [q1, q3] = np.split(res[0], 2)
    [q2, q4] = np.split(res[1], 2)
    return q1, q2, q3, q4


def join_matrix(q1, q2, q3, q4):
    q1, q2, q3, q4 = q4, q3, q2, q1
    return np.append(np.append(q1, q2, axis=1), np.append(q3, q4, axis=1), axis=0)


def img_scram(img):
    if img.shape == (1, 1):
        return img
    q1, q2, q3, q4 = split_quadrants(img)
    q1 = img_scram(q1)
    q2 = img_scram(q2)
    q3 = img_scram(q3)
    q4 = img_scram(q4)
    return np.rot90(join_matrix(q1, q2, q3, q4))


def img_rescram(img):
    if img.shape == (1, 1):
        return img
    q1, q2, q3, q4 = split_quadrants(img)
    q1 = img_rescram(q1)
    q2 = img_rescram(q2)
    q3 = img_rescram(q3)
    q4 = img_rescram(q4)
    return np.rot90(join_matrix(q1, q2, q3, q4), -1)


def scram(image, iterations=10):
    image = img_scram(image)
    Image.fromarray(image).save("Semi.png")
    height, width = image.shape
    new_image = np.zeros_like(image)
    for _ in range(iterations):
        for y in range(height):
            for x in range(width):
                new_x = (3 * x + 4 * y) % width
                new_y = (2 * x + 3 * y) % height
                new_image[new_y, new_x] = image[y, x]
        image = new_image.copy()
    return Image.fromarray(new_image)


def unscram(image, iterations=10):
    height, width = image.shape
    new_image = np.zeros_like(image)
    for _ in range(iterations):
        for y in range(height):
            for x in range(width):
                new_x = (3 * x - 4 * y) % width
                new_y = (-2 * x + 3 * y) % height
                new_image[new_y, new_x] = image[y, x]
        image = new_image.copy()
    new_image = img_rescram(new_image)
    return Image.fromarray(new_image)

def chaotic_map(x0, y0, a, b,c, n):
    chaotic_sequence = []
    x, y = x0, y0
    for _ in range(1000):
        x_new = b*y + a - x**2 + c*x**2
        y_new =  x
        x, y = x_new, y_new
    for _ in range(n):
        x_new =b*y + a - x**2 + c*x**2
        y_new = x
        x, y = x_new, y_new
        chaotic_sequence.append(x)
    return chaotic_sequence


def generate_key_and_nonce(x0,y0):
    a, b,c = 1.4, 0.3,0.05
    sequence = chaotic_map(x0, y0, a, b,c, 66)
    scaled_sequence = [int((x*10**3)%256) for x in sequence]
    byte_data = bytes(scaled_sequence)
    key = byte_data[:32]
    nonce = byte_data[32:56]
    return key, nonce

def encrypt_image(image_path, x,y):
    key,nonce = generate_key_and_nonce(x,y)
    original_image = Image.open(image_path)
    image_data = np.array(original_image)
    cipher = ChaCha20.new(key=key, nonce=nonce)
    encrypted_data = cipher.encrypt(image_data.tobytes())
    encrypted_image = Image.frombytes(original_image.mode, original_image.size, encrypted_data)

    return encrypted_image
  
def decrypt_image(enc_path, x, y):
    encrypted_image = Image.open(enc_path)
    image_data = np.array(encrypted_image)
    key,nonce = generate_key_and_nonce(x,y)
    cipher = ChaCha20.new(key=key, nonce=nonce)
    decrypted_data = cipher.decrypt(image_data.tobytes())
    decrypted_image = Image.frombytes(encrypted_image.mode, encrypted_image.size, decrypted_data)
    return decrypted_image